package com.wily.profserv.logreader;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.io.RandomAccessFile;

/**
 * @author Sylvain Gibassier
 * @copyright 2004-2005 Wily Technology, Inc.  All rights reserved.
 */

public class LogReader {
    static final private Logger logger = LogFactory.getLogger();

    public static void main(String[] args) {
        main(args, System.out);
    }

    public static void main(String[] args, PrintStream ps) {
       
        if (args.length > 3 || args.length < 2) {
            logger.error("Invalid configuration. LogReader needs 2 or 3 arguments");
            return;
        }
                
        File logFileToMonitor = new File(args[0]);
        
        String propertiesFileName = args[1];
        
        // lineProcessorClassName
        String lineProcessorClassName = "com.wily.profserv.logreader.SimpleMatcherLineProcessor";
        
        if (args.length == 3) {
            lineProcessorClassName = args[2];
        }
        
        logger.info("Starting...");
        
        LineProcessor lineProcessor = null;
        try {
            lineProcessor = LineProcessorFactory.getLineProcessor(lineProcessorClassName,ps,propertiesFileName);
        } catch (LineProcessorInitializationFailedException e) {
            logger.error(e.getMessage());
            return;
        }

        while (true) {
            // give the lineProcessor the opportunity to clear its fault status if necessary
            lineProcessor.tick();
            
            // check whether the file exists and is a file
            if (logFileToMonitor.exists() && logFileToMonitor.isFile()) {
                readAndProcessAsMuchAsPossible(logFileToMonitor,lineProcessor);
            } else {
                // the file either does not exist or it is not a file
                // log a warning message
                if (!logFileToMonitor.exists()) {
                    logger.warn("Log file \"" + logFileToMonitor.getAbsolutePath()
                            + "\" does not exist");
                } else {
                    logger.warn("Filesystem object \"" + logFileToMonitor.getAbsolutePath()
                            + "\" is not a file");
                }
                wait30Seconds();
            }
        }

    }

    static public void wait30Seconds() {
        try {
            Thread.sleep(30000);
        } catch (InterruptedException e) {
            // ignore
        }
    }

    static public void wait4Seconds() {
        try {
            Thread.sleep(4000);
        } catch (InterruptedException e) {
            // ignore
        }
    }

    static public void readAndProcessAsMuchAsPossible(File file, LineProcessor lineProcessor) {
        RandomAccessFile raf = null;
        StringBuffer cacheLine = new StringBuffer(3000);
        long position = 0;
        long fileLength = 0;
        byte[] buffer = new byte[1024];

        try {
            raf = new RandomAccessFile(file, "r");
            fileLength = raf.length();
            // position the cursor at the end of the file
            raf.seek(fileLength);
            position = fileLength;
            
            // the following 4 variables allows us to detect whether the underlying file
            // has been deleted and recreated after
            int numberOfWaits = 0;
            long cachedFsFileLength = -1;
            long cachedModificationTimestamp = -1;
            boolean fileDeleted = false;
            
            while (!fileDeleted && file.isFile() && file.exists()) {
                // give the lineProcessor the opportunity to clear its fault status if necessary
                lineProcessor.tick();
                
                long newFileLength = raf.length();
                
                // the 2 following variables allows us to detect whether the underlying file
                // has been deleted and recreated immediately after
                long fsFileLength = file.length();
                long modificationTimestamp = file.lastModified();
                
                if (newFileLength < fileLength) {
                    // reset the number of waits
                    numberOfWaits = 0;
                    // the file has been cut by the logging process
                    // starting from the beginning
                    position = 0;
                    fileLength = newFileLength;
                    cacheLine = new StringBuffer(3000);
                    raf.seek(0);
                } else if (newFileLength > fileLength) {
                    // the logging process has appended something to the file
                    fileLength = newFileLength;
                    long numberOfBytesTilEnd = fileLength - position;
                    // we want to read at most 1024 bytes
                    int numberOfBytesToRead = numberOfBytesTilEnd < 1024 ? (int) numberOfBytesTilEnd : 1024;

                    if (numberOfBytesToRead > 0) {
                        // reset the number of waits
                        numberOfWaits = 0;
                        
                        int bytesRead = raf.read(buffer, 0, numberOfBytesToRead);
                        
                        if (bytesRead > 0) {
                            position = position + bytesRead;
                            boolean skipLF = false;
                            endOfLineFindingLoop: for (int j = 0; j < bytesRead; j++) {
                                byte b = buffer[j];
                                if (skipLF) {
                                    skipLF = false;
                                    if (b == '\n') {
                                        continue endOfLineFindingLoop;
                                    }
                                }
                                switch (b) {
                                case '\r':
                                    skipLF = true;
                                case '\n': /* Fall through */
                                    // got a full line
                                    // send it to get processed
                                    lineProcessor.process(cacheLine.toString());
                                    cacheLine = new StringBuffer(3000);
                                    break;
                                default:
                                    cacheLine.append((char) b);
                                    break;
                                }
                            }

                        }
                    } else {
                        // nothing new in the log file
                        
                        // monitor if the file has been deleted and recreated immediately after
                        // RandomAccessFile does not detect that
                        // so we are checking whether the size
                        // or the modification timestamp of the file
                        // has been modified and the modification has not been picked
                        // up by the RandomAccessFile
                        if (numberOfWaits == 0) {
                            cachedFsFileLength = fsFileLength;
                            cachedModificationTimestamp = modificationTimestamp;
                        } else {
                            if (cachedFsFileLength != fsFileLength || cachedModificationTimestamp != modificationTimestamp) {
                                fileDeleted = true;
                                logger.warn("Log file \"" + file.getAbsolutePath() + "\" was deleted");
                            }
                        }
                        
                        numberOfWaits++;
                        if (numberOfWaits == Integer.MAX_VALUE) {
                            numberOfWaits = 1;
                        }
                        // let's wait!
                        wait4Seconds();
                    }
                } else {
                    // nothing new in the log file
                    
                    // monitor if the file has been deleted and recreated immediately after
                    // RandomAccessFile does not detect that
                    // so we are checking whether the size
                    // or the modification timestamp of the file
                    // has been modified and the modification has not been picked
                    // up by the RandomAccessFile
                    if (numberOfWaits == 0) {
                        cachedFsFileLength = fsFileLength;
                        cachedModificationTimestamp = modificationTimestamp;
                    } else {
                        if (cachedFsFileLength != fsFileLength || cachedModificationTimestamp != modificationTimestamp) {
                            fileDeleted = true;
                            logger.warn("Log file \"" + file.getAbsolutePath() + "\" was deleted");
                        }
                    }
                    
                    numberOfWaits++;
                    if (numberOfWaits == Integer.MAX_VALUE) {
                        numberOfWaits = 1;
                    }
                    // let's wait!
                    wait4Seconds();
                }
            }
        } catch (FileNotFoundException e) {
            // ignore: it will be detected in the main() loop anyway
        } catch (Throwable e) {
            logger.error(e.getMessage(), e);
        } finally {
            if (raf != null) {
                try {
                    raf.close();
                } catch (IOException e1) {
                    // ignore
                }
            }
        }
    }
}